@@ -41,7 +41,7 @@ try:
 except ImportError:
     PrettyPrinter = Any
 
-if TYPE_CHECKING:
+if TYPE_CHECKING:  # pragma: no cover
     if sys.version_info >= (3, 10):
         from typing import TypeGuard  # pylint: disable=no-name-in-module
     else:
@@ -63,6 +63,8 @@ class SliderKwargs(abc.Mapping):
     Sliders can be defined in :func:`~mpl_interactions.pyplot.interactive_plot`
     through :term:`kwargs <python:keyword argument>`. This wrapper class can be
     used for that.
+
+    .. automethod:: __getitem__
     """
 
     def __init__(
@@ -140,13 +142,13 @@ class SliderKwargs(abc.Mapping):
 
     def __repr__(self) -> str:
         return (
-            f"{self.__class__.__name__}("
+            f"{type(self).__name__}("
             f"sliders={self._sliders}, "
             f"arg_to_symbol={self._arg_to_symbol})"
         )
 
     def _repr_pretty_(self, p: PrettyPrinter, cycle: bool) -> None:
-        class_name = self.__class__.__name__
+        class_name = type(self).__name__
         if cycle:
             p.text(f"{class_name}(...)")
         else:
@@ -254,7 +256,7 @@ def _merge_args_kwargs(
     """
     output_dict = {}
     for arg in args:
-        if not isinstance(arg, dict):
+        if not isinstance(arg, abc.Mapping):
             raise TypeError("Positional arguments have to be of type dict")
         output_dict.update(arg)
     output_dict.update(kwargs)
@@ -303,7 +305,7 @@ def create_slider(symbol: sp.Symbol) -> "Slider":
     >>> create_slider(sp.Symbol("n0", integer=True))
     IntSlider(value=0, description='\\(n_{0}\\)')
     """
-    description = fR"\({sp.latex(symbol)}\)"
+    description = Rf"\({sp.latex(symbol)}\)"
     if symbol.is_integer:
         return IntSlider(description=description)
     return FloatSlider(description=description)
